@@ -1,12 +1,13 @@
 """
 Helpers for loading resources, managing specs, constants, etc.
 
-Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
+Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 SPDX-License-Identifier: MIT-0
 """
 import sys
 import fnmatch
 import json
+import hashlib
 import os
 import datetime
 import logging
@@ -18,7 +19,7 @@ import six
 from cfnlint.decode.node import dict_node, list_node, str_node
 from cfnlint.data import CloudSpecs
 try:
-    from urllib.request import urlopen
+    from urllib.request import urlopen, Request
 except ImportError:
     from urllib2 import urlopen
 try:
@@ -34,6 +35,7 @@ else:
 LOGGER = logging.getLogger(__name__)
 
 SPEC_REGIONS = {
+    'af-south-1': 'https://cfn-resource-specifications-af-south-1-prod.s3.af-south-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json',
     'ap-east-1': 'https://cfn-resource-specifications-ap-east-1-prod.s3.ap-east-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json',
     'ap-northeast-1': 'https://d33vqc0rt9ld30.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json',
     'ap-northeast-2': 'https://d1ane3fvebulky.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json',
@@ -46,6 +48,7 @@ SPEC_REGIONS = {
     'cn-northwest-1': 'https://cfn-resource-specifications-cn-northwest-1-prod.s3.cn-northwest-1.amazonaws.com.cn/latest/gzip/CloudFormationResourceSpecification.json',
     'eu-central-1': 'https://d1mta8qj7i28i2.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json',
     'eu-north-1': 'https://diy8iv58sj6ba.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json',
+    'eu-south-1': 'https://cfn-resource-specifications-eu-south-1-prod.s3.eu-south-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json',
     'eu-west-1': 'https://d3teyb21fexa9r.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json',
     'eu-west-2': 'https://d1742qcu2c1ncx.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json',
     'eu-west-3': 'https://d2d0mfegowb3wk.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json',
@@ -76,32 +79,6 @@ REGEX_DYN_REF_SSM = re.compile(r'^.*{{resolve:ssm:[a-zA-Z0-9_\.\-/]+:\d+}}.*$')
 REGEX_DYN_REF_SSM_SECURE = re.compile(r'^.*{{resolve:ssm-secure:[a-zA-Z0-9_\.\-/]+:\d+}}.*$')
 
 
-AVAILABILITY_ZONES = [
-    'ap-east-1a', 'ap-east-1b', 'ap-east-1c',
-    'ap-northeast-1a', 'ap-northeast-1b', 'ap-northeast-1c', 'ap-northeast-1d',
-    'ap-northeast-2a', 'ap-northeast-2b', 'ap-northeast-2c',
-    'ap-northeast-3a',
-    'ap-south-1a', 'ap-south-1b', 'ap-south-1c',
-    'ap-southeast-1a', 'ap-southeast-1b', 'ap-southeast-1c',
-    'ap-southeast-2a', 'ap-southeast-2b', 'ap-southeast-2c',
-    'ca-central-1a', 'ca-central-1b', 'ca-central-1d',
-    'cn-north-1a', 'cn-north-1b',
-    'cn-northwest-1a', 'cn-northwest-1b', 'cn-northwest-1c',
-    'eu-central-1a', 'eu-central-1b', 'eu-central-1c',
-    'eu-north-1a', 'eu-north-1b', 'eu-north-1c',
-    'eu-west-1a', 'eu-west-1b', 'eu-west-1c',
-    'eu-west-2a', 'eu-west-2b', 'eu-west-2c',
-    'eu-west-3a', 'eu-west-3b', 'eu-west-3c',
-    'me-south-1a', 'me-south-1b', 'me-south-1c',
-    'sa-east-1a', 'sa-east-1b', 'sa-east-1c',
-    'us-east-1a', 'us-east-1b', 'us-east-1c', 'us-east-1d', 'us-east-1e', 'us-east-1f',
-    'us-east-2a', 'us-east-2b', 'us-east-2c',
-    'us-gov-east-1a', 'us-gov-east-1b', 'us-gov-east-1c',
-    'us-gov-west-1a', 'us-gov-west-1b', 'us-gov-west-1c',
-    'us-west-1a', 'us-west-1b', 'us-west-1c',
-    'us-west-2a', 'us-west-2b', 'us-west-2c', 'us-west-2d', 'us-west-2-lax-1a',
-]
-
 FUNCTIONS = [
     'Fn::Base64', 'Fn::GetAtt', 'Fn::GetAZs', 'Fn::ImportValue',
     'Fn::Join', 'Fn::Split', 'Fn::FindInMap', 'Fn::Select', 'Ref',
@@ -130,27 +107,27 @@ PSEUDOPARAMS = [
 ]
 
 LIMITS = {
-    'mappings': {
-        'number': 100,
-        'attributes': 64,
+    'Mappings': {
+        'number': 200,
+        'attributes': 200,
         'name': 255  # in characters
     },
-    'outputs': {
-        'number': 60,
+    'Outputs': {
+        'number': 200,
         'name': 255,  # in characters
         'description': 1024  # in bytes
     },
-    'parameters': {
-        'number': 60,
+    'Parameters': {
+        'number': 200,
         'name': 255,  # in characters
         'value': 4096  # in bytes
     },
-    'resources': {
-        'number': 200,
+    'Resources': {
+        'number': 500,
         'name': 255  # in characters
     },
     'template': {
-        'body': 460800,  # in bytes
+        'body': 1000000,  # in bytes
         'description': 1024  # in bytes
     },
     'threshold': 0.9,  # for rules about approaching the other limit values
@@ -166,10 +143,144 @@ valid_snapshot_types = [
     'AWS::Redshift::Cluster'
 ]
 
+VALID_PARAMETER_TYPES_SINGLE = [
+    'AWS::EC2::AvailabilityZone::Name',
+    'AWS::EC2::Image::Id',
+    'AWS::EC2::Instance::Id',
+    'AWS::EC2::KeyPair::KeyName',
+    'AWS::EC2::SecurityGroup::GroupName',
+    'AWS::EC2::SecurityGroup::Id',
+    'AWS::EC2::Subnet::Id',
+    'AWS::EC2::VPC::Id',
+    'AWS::EC2::Volume::Id',
+    'AWS::Route53::HostedZone::Id',
+    'AWS::SSM::Parameter::Name',
+    'Number',
+    'String',
+    'AWS::SSM::Parameter::Value<AWS::EC2::AvailabilityZone::Name>',
+    'AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>',
+    'AWS::SSM::Parameter::Value<AWS::EC2::Instance::Id>',
+    'AWS::SSM::Parameter::Value<AWS::EC2::KeyPair::KeyName>',
+    'AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::GroupName>',
+    'AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::Id>',
+    'AWS::SSM::Parameter::Value<AWS::EC2::Subnet::Id>',
+    'AWS::SSM::Parameter::Value<AWS::EC2::VPC::Id>',
+    'AWS::SSM::Parameter::Value<AWS::EC2::Volume::Id>',
+    'AWS::SSM::Parameter::Value<AWS::Route53::HostedZone::Id>',
+    'AWS::SSM::Parameter::Value<AWS::SSM::Parameter::Name>',
+    'AWS::SSM::Parameter::Value<Number>',
+    'AWS::SSM::Parameter::Value<String>',
+]
 
-def get_url_content(url):
+VALID_PARAMETER_TYPES_LIST = [
+    'CommaDelimitedList',
+    'List<AWS::EC2::AvailabilityZone::Name>',
+    'List<AWS::EC2::Image::Id>',
+    'List<AWS::EC2::Instance::Id>',
+    'List<AWS::EC2::SecurityGroup::GroupName>',
+    'List<AWS::EC2::SecurityGroup::Id>',
+    'List<AWS::EC2::Subnet::Id>',
+    'List<AWS::EC2::VPC::Id>',
+    'List<AWS::EC2::Volume::Id>',
+    'List<AWS::Route53::HostedZone::Id>',
+    'List<Number>',
+    'List<String>',
+    'AWS::SSM::Parameter::Value<CommaDelimitedList>',
+    'AWS::SSM::Parameter::Value<List<AWS::EC2::AvailabilityZone::Name>>',
+    'AWS::SSM::Parameter::Value<List<AWS::EC2::Image::Id>>',
+    'AWS::SSM::Parameter::Value<List<AWS::EC2::Instance::Id>>',
+    'AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::GroupName>>',
+    'AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::Id>>',
+    'AWS::SSM::Parameter::Value<List<AWS::EC2::Subnet::Id>>',
+    'AWS::SSM::Parameter::Value<List<AWS::EC2::VPC::Id>>',
+    'AWS::SSM::Parameter::Value<List<AWS::EC2::Volume::Id>>',
+    'AWS::SSM::Parameter::Value<List<AWS::Route53::HostedZone::Id>>',
+    'AWS::SSM::Parameter::Value<List<Number>>',
+    'AWS::SSM::Parameter::Value<List<String>>',
+]
+
+VALID_PARAMETER_TYPES = VALID_PARAMETER_TYPES_SINGLE + VALID_PARAMETER_TYPES_LIST
+
+class RegexDict(dict):
+
+    def __getitem__(self, item):
+        possible_items = {k: v for k, v in self.items() if re.match(k, item)}
+        if not possible_items:
+            raise KeyError
+        longest_match = sorted(possible_items.keys(), key=len)[-1]
+        return possible_items[longest_match]
+
+    def __contains__(self, item):
+        for k in self.keys():
+            if re.match(k, item):
+                return True
+        return False
+
+    def get(self, key, default=None):
+        try:
+            return self[key]
+        except KeyError:
+            return default
+
+def get_metadata_filename(url):
+    """Returns the filename for a metadata file associated with a remote resource"""
+    caching_dir = os.path.join(os.path.dirname(__file__), 'data', 'DownloadsMetadata')
+    encoded_url = hashlib.sha256(url.encode()).hexdigest()
+    metadata_filename = os.path.join(caching_dir, encoded_url + '.meta.json')
+
+    return metadata_filename
+
+def url_has_newer_version(url):
+    """Checks to see if a newer version of the resource at the URL is available
+    Always returns true if using Python2.7 due to lack of HEAD request support,
+    or if we have no caching information for the local version of the resource
+    """
+    metadata_filename = get_metadata_filename(url)
+
+    # Load in the cache
+    metadata = load_metadata(metadata_filename)
+
+    # Etag is a caching identifier used by S3 and Cloudfront
+    if 'etag' in metadata:
+        cached_etag = metadata['etag']
+    else:
+        # If we don't know the etag of the local version, we should force an update
+        return True
+
+    # Need to wrap this in a try, as URLLib2 in Python2 doesn't support HEAD requests
+    try:
+        # Make an initial HEAD request
+        req = Request(url, method='HEAD')
+        res = urlopen(req)
+
+    except NameError:
+        # We should force an update
+        return True
+
+    # If we have an ETag value stored and it matches the returned one,
+    # then we already have a copy of the most recent version of the
+    # resource, so don't bother fetching it again
+    if cached_etag and res.info().get('ETag') and cached_etag == res.info().get('ETag'):
+        LOGGER.debug('We already have a cached version of url %s with ETag value of %s', url, cached_etag)
+        return False
+
+    # The ETag value of the remote resource does not match the local one, so a newer version is available
+    return True
+
+def get_url_content(url, caching=False):
     """Get the contents of a spec file"""
+
     res = urlopen(url)
+
+    if caching and res.info().get('ETag'):
+        metadata_filename = get_metadata_filename(url)
+        # Load in all existing values
+        metadata = load_metadata(metadata_filename)
+        metadata['etag'] = res.info().get('ETag')
+        metadata['url'] = url # To make it obvious which url the Tag relates to
+        save_metadata(metadata, metadata_filename)
+
+    # Continue to handle the file download normally
     if res.info().get('Content-Encoding') == 'gzip':
         buf = BytesIO(res.read())
         f = gzip.GzipFile(fileobj=buf)
@@ -180,6 +291,25 @@ def get_url_content(url):
     return content
 
 
+def load_metadata(filename):
+    """Get the contents of the download metadata file"""
+    metadata = {}
+    if os.path.exists(filename):
+        with open(filename, 'r') as metadata_file:
+            metadata = json.load(metadata_file)
+    return metadata
+
+
+def save_metadata(metadata, filename):
+    """Save the contents of the download metadata file"""
+    dirname = os.path.dirname(filename)
+    if not os.path.exists(dirname):
+        os.mkdir(dirname)
+
+    with open(filename, 'w') as metadata_file:
+        json.dump(metadata, metadata_file)
+
+
 def load_resource(package, filename='us-east-1.json'):
     """Load CloudSpec resources
         :param filename: filename to load
@@ -189,7 +319,7 @@ def load_resource(package, filename='us-east-1.json'):
 
 
 RESOURCE_SPECS = {}
-
+REGISTRY_SCHEMAS = []
 
 def merge_spec(source, destination):
     """ Recursive merge spec dict """
